import os
from logging import getLogger

from abeja.common.file_helpers import extract_zipfile
from abeja.exceptions import BadRequest, Unauthorized, Forbidden, NotFound, MethodNotAllowed, \
    ResourceNotFound
from abeja.train.api.client import APIClient
from abeja.train.statistics import Statistics


class Client:
    """A High-Level client for Training API

    .. code-block:: python

        from abeja.train import Client

        client = Client()

    """

    ORGANIZATION_ID = os.environ.get('ABEJA_ORGANIZATION_ID')
    JOB_DEFINITION_NAME = os.environ.get('TRAINING_JOB_DEFINITION_NAME')
    TRAINING_JOB_ID = os.environ.get('TRAINING_JOB_ID')

    def __init__(self) -> None:
        self.api = APIClient()
        self.logger = getLogger('train-api')

    def download_training_result(self, training_job_id: str, path: str=None) -> None:
        """download training artifact that includes model file or logs.
        Training artifact itself is a zip file, this function extracts it.

        API reference: GET /organizations/<organization_id>/training/definitions/<job_definition_name>/jobs/<training_job_id>/result

        Request Syntax:
            .. code-block:: python

                training_job_id = '1600000000000'
                client.download_training_result(training_job_id)

        Params:
            - **training_job_id** (str): training job id
            - **path** (str): target path where result are extracted **[optional]**

        Raises:
            - abeja.exceptions.BadRequest
            - abeja.exceptions.Unauthorized
            - abeja.exceptions.Forbidden
            - abeja.exceptions.NotFound
            - abeja.exceptions.MethodNotAllowed
            - abeja.exceptions.InternalServerError
            - json.JSONDecodeError
            - PermissionError
            - ValueError
            - IOError
        """
        response = self.api.get_training_result(organization_id=Client.ORGANIZATION_ID,
                                                job_definition_name=Client.JOB_DEFINITION_NAME,
                                                training_job_id=training_job_id)
        try:
            artifact_uri = response['artifacts']['complete']['uri']
        except KeyError:
            raise ResourceNotFound(
                'training result for training_id {} is not found'.format(training_job_id))
        content = self._get_content(artifact_uri)
        extract_zipfile(content, path=path)

    def _get_content(self, url: str) -> bytes:
        """get binary content from given url.

        :param url:
        :return:
        :raises: abeja.exceptions.BadRequest
                 abeja.exceptions.Unauthorized
                 abeja.exceptions.Forbidden
                 abeja.exceptions.NotFound
                 abeja.exceptions.MethodNotAllowed
                 abeja.exceptions.InternalServerError
                 json.JSONDecodeError
        """
        res = self.api._connection.request('GET', url)
        return res.content

    def update_statistics(self, statistics: Statistics) -> None:
        """ Notify a job statistics for ABEJA Platform.

        API reference: POST /organizations/<organization_id>/training/definitions/<job_definition_name>/jobs/<training_job_id>/statistics

        Request Syntax:
            .. code-block:: python

                from abeja.train import Client
                from abeja.train.statistics import Statistics as ABEJAStatistics

                client = Client()

                statistics = ABEJAStatistics(num_epochs=10, epoch=1)
                statistics.add_stage(name=ABEJAStatistics.STAGE_TRAIN, accuracy=90.0, loss=0.10)
                statistics.add_stage(name=ABEJAStatistics.STAGE_VALIDATION, accuracy=75.0, loss=0.07)

                client.update_statistics(statistics)

        Params:
            - **statistics** (:class:`abeja.train.statistics.Statistics`): job statistics to nofity

        Returns:
            None
        """
        if not statistics or not statistics.get_statistics():
            self.logger.warning('empty statistics found.')
            return

        try:
            response = self.api.update_statistics(organization_id=Client.ORGANIZATION_ID,
                                                  job_definition_name=Client.JOB_DEFINITION_NAME,
                                                  training_job_id=Client.TRAINING_JOB_ID,
                                                  statistics=statistics.get_statistics())
            self.logger.info('update_statistics result: %s', response)
        except (BadRequest, Unauthorized, Forbidden, NotFound, MethodNotAllowed) as e:
            self.logger.warning('update_statistics result was {}.'.format(str(e)))
        except Exception:
            self.logger.exception('update_statistics result was unexpected error:')
